/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  codecan.hpp
///  \brief Implements the canonization algorithm for linear codes over finite fields and chain rings with 4 elements.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef CODECAN_H_
#define CODECAN_H_

#include "gennode.hpp"

#include "helpfunction.hpp"
#include "actingElement.hpp"
#include <sstream>

#define GNAME "G" ///< Will be used for the communication with the computer algebra system.
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
/// \class CodeCan
/// \brief The canonization algorithm for linear codes over chain rings with 4 elements.
///
/// This implements the algorithm described in <em>T. Feulner, Canonization of linear codes over \f$\mathbb{Z}_4\f$, AMC, submitted</em> and
/// <em> T. Feulner; The automorphism groups of  linear codes and canonical representatives of their semilinear isometry classes, AMC 3(4), pp. 363 - 383
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
template<typename RING, AlgTypeE ALGTYPE>
class CodeCan;

template<AlgTypeE ALGTYPE>
class CodeCanStarter
{
public:
    /// \brief Start the canonization process with the \a root as the root node.
    /// \arg \c k0 The parameter of the type of the generator matrix described by root.
    /// \arg \c root The root node of the backtrack tree.
    ///
    /// We first search for columns that are multiples of each other (under multiplicaton by units).
    /// If all columns are linearly independent we start the canonization process immediately.
    /// Otherwise, we build a local projective matrix, only containing one representative and start with this matrix as the root node.
    /// The columns are partitioned according to the multiplicity of their occurrence.
    /// Afterwards, the resulting group elements are blown up to the initial length.
    /// \see CodeCan::canonize
    template<typename RING>
    static void start_generator_matrix(CodeCan<RING, ALGTYPE>& canonizer, unsigned int k0,
            const Node<RING, ALGTYPE>& root, bool only_automs);//, bool minWeightOfDual, unsigned int givenLeastMinWeight);

    /// \brief This function decides, whether to start the canonization process for Gamma or a generator matrix of the dual.
    template<typename RING>
    static void start(CodeCan<RING, ALGTYPE>& canonizer, const Matrix<RING>& Gamma, bool only_automs);

    template<typename RING>
    static void advanced_start(unsigned int k0, CodeCan<RING, ALGTYPE>& canonizer,  const Node<RING, ALGTYPE>& root, bool only_automs, const std::vector<Permutation>& known_automs);

private:
    template<typename RING>
    static void addTrivialMults(const Matrix<RING>& mat, CodeCan<RING, ALGTYPE>& canonizer);

};

template<typename RING, AlgTypeE ALGTYPE>
class CodeCan
{
private:
    //    bool m_minWeightStop; // the canonization was not started (only needed for classification)
    Node<RING, ALGTYPE> m_root; ///< The root node of the backtrack tree.
    Matrix<RING> m_GammaBest; ///< The best candidate for the canonical representative.
    InnerStab<RING, ALGTYPE> m_innerStab2GammaBest; ///< The inner stabilizer of GammaBest.
    std::vector<ActingElement<RING, ALGTYPE> > m_autom; ///< Generators of the automorphism group (or the known part).
    My_Long_Int m_automOrderProjective, m_automOrderFull; ///< The order of the automorphism group.
    AbstractCASModule* m_cas; ///< A computer algebra system which calculates the labelled branching.
    bool m_deleteCAS; ///< Should we delete the computer algebra system after finishing.

    ActingElement<RING, ALGTYPE> m_transporter; ///< The group element we have to apply to \a m_root to get \a m_GammaBest.
    ActingElement<RING, ALGTYPE> m_transporterInverse; ///< The group element we have to apply to \a m_GammaBest to get \a m_root.

    /// Fill the remaining part of the acting elements, during the backtracking we only fill the permutations.
    void fillAutoms(unsigned int k0);

    /// Fill the remaining part of the acting elements, during the backtracking we only fill the permutations.
    void fillTransporter(unsigned int k0);

    /// Start the canonization process.
    bool canonize(unsigned int k0, bool only_automs, GenNode<RING, ALGTYPE>* p_backtrack=NULL, const std::vector<Permutation>* const known_automs=NULL);//, bool minWeightOfDual, unsigned int givenLeastMinWeight);

    /// Add a new automorphism generator and replace the labelled branching.
    void addAutomorphism(int n, GenNode<RING, ALGTYPE>* const p_backtrack, const Permutation& new_autom);

    friend class CodeCanStarter<ALGTYPE> ;

public:
    /// Constructor, starting GAP in the sage directory.
    CodeCan(char* sage_root);

    /// Constructor, passing a computer algebra system.
    CodeCan(AbstractCASModule* cas) :
        m_automOrderProjective(1), m_cas(cas), m_deleteCAS(false)
    {
    }
    ~CodeCan()
    {
        if (m_deleteCAS)
        {
            delete m_cas;
        }
    }

    //    bool get_minWeightStop(void) const
    //    {
    //        return m_minWeightStop;
    //    }

    /// Start the canonization process for \a Gamma.
    /// \see CodeCan::start
    //    void start_matrix(const Matrix<RING>& Gamma);

    /// Start the canonization process for a matrix \a Gamma which is stored in the file \a filename.
    /// \see CodeCan::start
    void start_file(char* filename);

    /// Get the order of the automorphism group.
    std::pair<My_Long_Int, My_Long_Int> getAutomOrder(void) const;

    /// Get the order of the automorphism group.
    //    const mpz_t& getAutomOrder(mpz_t& res) const;

    /// Get the transporter element, i.e. the group element \a m_transporter with m_transporter*m_root = m_GammaBest
    const ActingElement<RING, ALGTYPE>& getTransporter(void) const
                                                                            {
        return m_transporter;
                                                                            }

    /// Get the inverse of the transporter element, i.e. the group element \a m_transporterInverse with m_transporter* m_GammaBest = m_root
    const ActingElement<RING, ALGTYPE>& getTransporterInverse(void) const
                                                                            {
        return m_transporterInverse;
                                                                            }

    /// Get the generators of the automorphism group of m_root.
    const std::vector<ActingElement<RING, ALGTYPE> >& getAutoms(void) const
                                                                            {
        return m_autom;
                                                                            }

    /// Get the canonical representative of the orbit of equivalent matrices.
    const Matrix<RING>& getGammaBest(void) const
                                                                            {
        return m_GammaBest;
                                                                            }

    AbstractCASModule* getCAS(void) const
                                                                            {
        return m_cas;
                                                                            }
};

template<typename RING, AlgTypeE ALGTYPE>
void CodeCan<RING, ALGTYPE>::start_file(char* filename)
{
    Matrix<RING> mat(filename);
    start(mat);
}

template<AlgTypeE ALGTYPE>
template<typename RING>
void CodeCanStarter<ALGTYPE>::start(CodeCan<RING, ALGTYPE>& canonizer, const Matrix<RING>& Gamma, bool only_automs)
{
    /// todo: Should I specialize this for a finite field in order to avoid the calculation of the check matrix?
    Matrix<RING> Delta;
    unsigned int k0 = Matrix<RING>::checkMatrix(Gamma, Delta); // I need to know the type for the chain rings Z_4 and F_2[X]/(X^2)

    unsigned int k = Gamma.getRows();
    if (Delta.getRows() < k)
    {
        // it is (likely) more efficient to work with the dual code
        if (Delta.getRows() == 0)
        {
            /// todo: catch this special, but stupid case
            std::cout << "The length of the code is zero, stopped calculation" << std::endl;
            exit(-1);
        }
        unsigned int my_k0 = Gamma.getCols() - k;
        Node<RING, ALGTYPE> root(my_k0, Delta);
        start_generator_matrix(canonizer, my_k0, root, only_automs);

        // transform the solution
        Matrix<RING>::checkMatrix(canonizer.m_GammaBest, canonizer.m_GammaBest);

        // in order to update the group elements we need some unique generator matrix of the code generated by Gamma
        Matrix<RING> G(Gamma), A;
        G.toOrderedBasis(true, NULL, &A);
        Matrix<RING> A_inv(A.inverse());

        canonizer.m_transporterInverse.m_A.identity(k);
        canonizer.m_transporterInverse.phi2multInverse();
        G = canonizer.m_transporterInverse * canonizer.m_GammaBest;
        Matrix<RING> B;
        G.toOrderedBasis(true, NULL, &B);
        canonizer.m_transporterInverse.m_A = A_inv * B;
        canonizer.m_transporter.phi2multInverse();
        canonizer.m_transporter.m_A = canonizer.m_transporterInverse.m_A.inverse();
        if (ALGTYPE == SEMILINEAR)
        {
            canonizer.m_transporter.m_A_applyOwnFrobenius();
        }

        for (unsigned int i = 0; i < canonizer.m_autom.size(); ++i)
        {
            canonizer.m_autom[i].phi2multInverse();
            canonizer.m_autom[i].m_A.identity(k);
            G = canonizer.m_autom[i] * Gamma;
            G.toOrderedBasis(true, NULL, &B);
            canonizer.m_autom[i].m_A = A_inv * B;
        }
    }
    else
    {
        // we work on the code itself
        Node<RING, ALGTYPE> root(k0, Gamma);
        start_generator_matrix(canonizer, k0, root, only_automs);
    }
}

template<AlgTypeE ALGTYPE>
template<typename RING>
void CodeCanStarter<ALGTYPE>::advanced_start(unsigned int k0, CodeCan<RING, ALGTYPE>& canonizer, const Node<RING, ALGTYPE>& root, bool only_automs, const std::vector<Permutation>& known_automs)
{
    canonizer.m_autom.clear();
    canonizer.m_root = root;
    GenNode<RING, ALGTYPE> backtrack(k0, root, only_automs);
    canonizer.canonize(k0, only_automs, &backtrack, &known_automs);
}

template<AlgTypeE ALGTYPE>
template<typename RING>
void CodeCanStarter<ALGTYPE>::addTrivialMults(const Matrix<RING>& mat, CodeCan<RING, ALGTYPE>& canonizer)
{
    unsigned int n = mat.getCols();
    unsigned int k = mat.getRows();
    std::vector<Element<RING> > gens_multGroup(RING::multGenerator());
    std::vector<unsigned int> heights = mat.heightSequence(0);
    std::vector<unsigned int>::const_iterator heightsIter = heights.begin();

    Vec<RING> phi(n, 1);
    for (unsigned int i = 0; i < n; ++i, ++heightsIter)
    {
        if (*heightsIter != 0)
        {
            //no unit is contained
            for (unsigned int j = 0; j < gens_multGroup.size(); ++j)
            {
                phi[i] = gens_multGroup[j];
                canonizer.m_autom.push_back(ActingElement<RING, ALGTYPE> (k, phi));
                phi[i] = 1;
            }
            canonizer.m_automOrderFull *= RING::my_multiplicativeGroupSize();
        }
    }
}

template<>
template<typename RING>
void CodeCanStarter<PLUSMINUSMONOMIAL>::addTrivialMults(const Matrix<RING>& mat, CodeCan<RING, PLUSMINUSMONOMIAL>& canonizer)
{
    Element<RING> minusOne( -Element<RING>(1) );
    if( minusOne.getIndex() == 1)
    {
        return;
    }

    std::vector<unsigned int> heights = mat.heightSequence(0);
    std::vector<unsigned int>::const_iterator heightsIter = heights.begin();
    unsigned int n = mat.getCols();
    unsigned int k = mat.getRows();
    Vec<RING> phi(n, 1);
    for (unsigned int i = 0; i < n; ++i, ++heightsIter)
    {
        if (*heightsIter != 0)
        {
            //no unit is contained
            phi[i] = minusOne;
            canonizer.m_autom.push_back(ActingElement<RING, PLUSMINUSMONOMIAL> (k, phi));
            phi[i] = 1;
            canonizer.m_automOrderFull *= 2;
        }
    }
}

template<>
template<typename RING>
void CodeCanStarter<PERMUTATIONAL>::addTrivialMults(const Matrix<RING>& mat, CodeCan<RING, PERMUTATIONAL>& canonizer)
{
    // there is nothing to do
}


template<AlgTypeE ALGTYPE>
template<typename RING>
void CodeCanStarter<ALGTYPE>::start_generator_matrix(CodeCan<RING, ALGTYPE>& canonizer, unsigned int k0, const Node<
        RING, ALGTYPE>& root, bool only_automs)//, bool minWeightOfDual, unsigned int givenLeastMinWeight)
        {
    //	std::cout<<root<<std::endl;
    canonizer.m_autom.clear();
    const Matrix<RING>& mat = root.getMatrix();
    unsigned int n = mat.getCols();
    unsigned int k = mat.getRows();
    Vec<RING> projMult(n, 1);
    std::vector<int> projPermList(n, -1);

    std::map<Vec<RING> , std::vector<unsigned int> > colToIndices;
    for (unsigned int i = 0; i < n; ++i)
    {
        Vec<RING> col = mat.getColumn(i);
        if(ALGTYPE!=PERMUTATIONAL)
        {
            projMult[i] = col.normalize(ALGTYPE==PLUSMINUSMONOMIAL);
        }
        colToIndices[col].push_back(i);
    }
    std::vector<std::vector<Vec<RING> > > colsByMultiplicity(n + 1);
    typename std::map<Vec<RING> , std::vector<unsigned int> >::const_iterator colToIndexIter(colToIndices.begin());
    typename std::map<Vec<RING> , std::vector<unsigned int> >::const_iterator colToIndexIterEnd(colToIndices.end());
    while (colToIndexIter != colToIndexIterEnd)
    {
        colsByMultiplicity[colToIndexIter->second.size()].push_back(colToIndexIter->first);
        ++colToIndexIter;
    }

    if (colsByMultiplicity[1].size() == n)
    { // there are no multiple columns....
        // we start the canonization process with root
        canonizer.m_root = root;
        canonizer.canonize(k0, only_automs);//, minWeightOfDual, givenLeastMinWeight);
        addTrivialMults(mat, canonizer);
        return;
    }

    // we start the canonization process with a projective version of root, were the columns are partitioned by the multiplicity of their occurence.
    std::vector<Vec<RING> > proj_entries;
    OrdPartition blocks;
    blocks.push_back(true);
    std::vector<std::vector<unsigned int> > new2oldIndices;
    unsigned int counter(0);
    for (unsigned int i = 1; i <= n; ++i)
    {
        for (unsigned int j = 0; j < colsByMultiplicity[i].size(); ++j)
        {
            const Vec<RING>& newCol = colsByMultiplicity[i][j];
            proj_entries.push_back(newCol);
            new2oldIndices.push_back(colToIndices[newCol]);
            blocks.push_back(false);

            std::vector<unsigned int>& old_cols = colToIndices[newCol];
            for (unsigned l = 0; l < old_cols.size(); ++l)
            {
                projPermList[counter] = old_cols[l];
                ++counter;
            }
        }
        blocks.back() = true;
    }
    blocks.pop_back(); // there is one to much
    canonizer.m_root = Node<RING, ALGTYPE> (k0, Matrix<RING> (proj_entries), &blocks);

//    std::cout<<canonizer.m_root<<std::endl;

    Permutation projPerm(projPermList), projPermInverse;
    Permutation::inv(projPerm, projPermInverse);
    ActingElement<RING, ALGTYPE> oldTransport(k, projMult.permute(projPermInverse), projPermInverse);
    ActingElement<RING, ALGTYPE> oldTransportInverse(k, projMult.multInverse(), projPerm);
    //
    //	cout<<oldTransport<<endl;
    //	cout<<oldTransportInverse<<endl;
    //	cout<<oldTransport*oldTransportInverse<<endl;

    canonizer.canonize(k0, only_automs);//, minWeightOfDual, givenLeastMinWeight);

    // blow up the result to the initial number of columns.
    std::vector<Vec<RING> > GammaBest_entries;
    for (unsigned int i = 0; i < canonizer.m_root.getCols(); ++i)
    {
        const Vec<RING>& column = canonizer.m_GammaBest.getColumn(i);
        for (unsigned int j = 0; j < new2oldIndices[i].size(); ++j)
        {
            GammaBest_entries.push_back(column);
        }
    }
    canonizer.m_GammaBest = Matrix<RING> (GammaBest_entries);

    canonizer.m_transporter.enlarge(n, new2oldIndices);
    canonizer.m_transporter *= oldTransport;
    canonizer.m_transporterInverse.enlarge(n, new2oldIndices);
    canonizer.m_transporterInverse = oldTransportInverse * canonizer.m_transporterInverse;
    for (unsigned int i = 0; i < canonizer.m_autom.size(); ++i)
    {
        ActingElement<RING, ALGTYPE>& aut = canonizer.m_autom[i];
        aut.enlarge(n, new2oldIndices);
        aut *= oldTransport;
        aut = oldTransportInverse * aut;
    }

    counter = 0;
    for (unsigned int i = 0; i < canonizer.m_root.getCols(); ++i)
    {
        unsigned int mySize = new2oldIndices[i].size();
        if (mySize > 1)
        {
            canonizer.m_automOrderFull *= help::factorial(mySize);
            Permutation a;
            a.transposition(counter, counter + 1);
            ActingElement<RING, ALGTYPE> aut(k, n, a);
            aut *= oldTransport;
            aut = oldTransportInverse * aut;
            canonizer.m_autom.push_back(aut);
            if (mySize > 2)
            {
                a.cycle(counter, counter + mySize - 1);
                aut = ActingElement<RING, ALGTYPE> (k, n, a);
                aut *= oldTransport;
                aut = oldTransportInverse * aut;
                canonizer.m_autom.push_back(aut);
            }
            counter += mySize;
        }
        else
        {
            ++counter;
        }
    }

    addTrivialMults(mat, canonizer);

        }

template<typename RING, AlgTypeE ALGTYPE>
bool CodeCan<RING, ALGTYPE>::canonize(unsigned int k0, bool only_automs, GenNode<RING, ALGTYPE>* p_backtrack, const std::vector<Permutation>* const known_automs)
//, bool minWeightOfDual, unsigned int givenLeastMinWeight)
{
    m_autom.clear();
    m_automOrderProjective = 1;
    int n = m_root.getCols();

    bool delete_p_backtrack(false);
    if(p_backtrack == NULL)
    {
        delete_p_backtrack = true;
        p_backtrack = new GenNode<RING, ALGTYPE> (k0, m_root, only_automs);
    }

    if(known_automs == NULL)
    {
        m_cas->initTrivialSn(GNAME, n);
    }
    else
    {
        m_autom.resize(known_automs->size());
        typename std::vector<ActingElement<RING, ALGTYPE> >::iterator automIter(m_autom.begin());
        for( std::vector<Permutation>::const_iterator knownIter(known_automs->begin()); knownIter<known_automs->end(); ++knownIter, ++automIter)
        {
            automIter->m_pi = *knownIter;
        }

        m_cas->buildGroup(GNAME, *known_automs, n);
        std::vector<int> father(n);
        m_cas->getFather(GNAME, father, m_automOrderProjective);
        p_backtrack->resetFather(father);
    }

    std::vector<unsigned int> type(1, k0);
    if (RING::chainlength() == 2)
    {
        type.push_back(m_root.getRows() - k0);
    }

    (*p_backtrack)->initMinWords(m_cas, type);

    int lvl = -1;
    for (p_backtrack->begin(); !p_backtrack->end(); p_backtrack->next(lvl))
    {
        lvl = p_backtrack->level();
//        		std::cout<<lvl<<std::endl;
//           		std::cout<<**p_backtrack<<std::endl;


        if (!p_backtrack->HasDiscreteBlocks())
        {
            // the actual node has not got a discrete partition, we have to split some block and backtrack further.
            ++lvl;
            continue;
        }

        if (!p_backtrack->doDiscreteCompletion())
        {
            continue; // the actual node gives not a canonical representative
        }

        if (p_backtrack->getBestInitialized())
        {
            // an automorphism
            p_backtrack->writeNode("Autom");
            // replace the father vector of the labelled branching.
            addAutomorphism(n, p_backtrack, m_transporterInverse.m_pi * p_backtrack->getPermutation());
        }
        else
        {
            // a new solution
            //            std::stringstream text("new Cand");
            //            text<<lvl;
            p_backtrack->writeNode("new Cand");
            // store all neccessary data
            m_transporter.m_pi = p_backtrack->getPermutation();
            Permutation::inv(m_transporter.m_pi, m_transporterInverse.m_pi);
            p_backtrack->setBestInitialized();
            (*p_backtrack)->saveAsBest(m_GammaBest, m_innerStab2GammaBest);

            //            if(secondMatch)
            //            {
            //                break;
            //            }else
            //            {
            //                secondMatch=true;
            //            }
        }
    }
    if( delete_p_backtrack)
    {
        delete p_backtrack;
    }

    // backtracking is done, fill the group elements (since only the permutational part is stored during the backtracking)
    fillTransporter(k0);
    fillAutoms(k0);
    return true;
}

template<typename RING, AlgTypeE ALGTYPE>
void CodeCan<RING, ALGTYPE>::addAutomorphism(int n, GenNode<RING, ALGTYPE>* const p_backtrack,
        const Permutation& new_autom)
        {
    m_autom.push_back(ActingElement<RING, ALGTYPE> (new_autom));
    m_cas->extendGroup(GNAME, new_autom, n);

    std::vector<int> father(n);
    m_cas->getFather(GNAME, father, m_automOrderProjective);

//    std::cout<<m_automOrder<<std::endl;

    p_backtrack->resetFather(father);
        }

template<typename RING, AlgTypeE ALGTYPE>
std::pair<My_Long_Int, My_Long_Int> CodeCan<RING, ALGTYPE>::getAutomOrder(void) const
{
    return std::pair<My_Long_Int, My_Long_Int>(m_automOrderFull, m_automOrderProjective);
}

//template<typename RING, AlgTypeE ALGTYPE>
//const mpz_t& CodeCan<RING, ALGTYPE>::getAutomOrder(mpz_t& res) const
//{
//#ifdef MPZ
//    mpz_set(res,  m_automOrder.get_mpz_t());
//#else
//    mpz_set_ui(res, m_automOrder);
//#endif
//    return res;
//}

template<typename RING, AlgTypeE ALGTYPE>
void CodeCan<RING, ALGTYPE>::fillTransporter(unsigned int k0)
{
    unsigned int n = m_GammaBest.getCols();
    unsigned int k = m_GammaBest.getRows();
    m_transporter.reset(k, n);
    InnerStab<RING, ALGTYPE> stab(k0, n);
    stab.fillActingElement(k0, m_root.getMatrix(), m_transporter, m_innerStab2GammaBest.m_fixedCols);
    m_transporterInverse = m_transporter.inverse();
}

template<typename RING, AlgTypeE ALGTYPE>
void CodeCan<RING, ALGTYPE>::fillAutoms(unsigned int k0)
{
    //	m_automOrder = m_cas->getOrder(GROUPNAME);
    unsigned int n = m_GammaBest.getCols();
    unsigned int k = m_GammaBest.getRows();
    ActingElement<RING, ALGTYPE> rootMinimizer;

    for (unsigned int i = 0; i < m_autom.size(); ++i)
    {
        m_autom[i].reset(k, n);
        m_autom[i].m_pi = m_transporter.m_pi * m_autom[i].m_pi; // we have to go via Gamma_best...
        InnerStab<RING, ALGTYPE> stab(k0, n);
        stab.fillActingElement(k0, m_root.getMatrix(), m_autom[i], m_innerStab2GammaBest.m_fixedCols);
        m_autom[i] = m_transporterInverse * m_autom[i];
    }
    m_automOrderFull = m_automOrderProjective;
    std::vector<ActingElement<RING, ALGTYPE> > remInnerStabGens;
    m_automOrderFull *= m_innerStab2GammaBest.getAutom(m_GammaBest, remInnerStabGens);
    for (unsigned int i = 0; i < remInnerStabGens.size(); ++i)
    {
        m_autom.push_back(m_transporterInverse * remInnerStabGens[i] * m_transporter);
    }
}

#endif /*CODECAN_H_*/
