/* This is  -*- C++ -*-  */
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
///  \file  partition.h
///  \brief Here, everything needed for partitioning sets is implemented.
///  \author Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  \date   2010-09-29
///
///  Copyright (C) 2010 Thomas Feulner <thomas.feulner@uni-bayreuth.de>
///  Distributed  under  the  terms  of  the  GNU  General  Public  License (GPL)
///  http://www.gnu.org/licenses/
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////


#ifndef PARTITION_H
#define PARTITION_H

#include <iostream>
#include <vector>
#include <list>

/////////////////////////////////////////////////////////////////////////////////////////
/// \class OrdPartition
/// \brief Just a vector<int> to represent canonical young groups.
///
/// A class to implement the column partition of a matrix.
/// The matrix will allways be permuted in such way, that the columns of same
/// type will be consecutive. Hence a block starting at position i and ending at
/// position j will be represented by true at i and j+1 (if j+1 <n) and false at the 
/// intermediate positions.
/// A matrix, which consists of only one block, has therefore only one true at
/// the position 0.
/////////////////////////////////////////////////////////////////////////////////////////

class OrdPartition:public std::vector<int>
{
	public:
		OrdPartition() : std::vector<int>(0) {}
		OrdPartition(int n) : std::vector<int>(n, 0) { if(n>0){this->front()=1;}}
		
		/// Init from a file.
		OrdPartition(int cols, const char* const filename);
		
		/// Compares the partition order array with the partition \a p  on positions >= \a lvl
		/// \return 
		///			- 0 if equal.
		/// 			- -1 if the i-th block of \a p is smaller and all preceeding ones are of equal size. 
		/// 			- 1 else.
		int comp(const OrdPartition& p, int lvl);
		
		/// Scan for a OrdPartition.
		bool scan(std::istream& in, unsigned int n);
		
		/// Print the ordered partition.		
		friend std::ostream & operator<< (std::ostream & os, const OrdPartition& p);
		
		std::vector<unsigned int> getBlocksizes(void) const;
		std::ostream& printBlocksizes(std::ostream& os) const;
};


/////////////////////////////////////////////////////////////////////////////////////////
/// \class UnordPartition
/// \brief Partition of the set n.
///
/// The partition is stored by an array \a index, which tells for
/// each i the index of the set to which i belongs.
/////////////////////////////////////////////////////////////////////////////////////////

class UnordPartition
{
	public:
		std::vector<int> m_index;				///< The set index for each i.
	
		/// Initialize with the discrete partition.
		UnordPartition(int n=0);
		
		void clear(void)
		{
			m_index.clear();
		}
		
		std::vector<std::vector<unsigned int> > getSets(void) const
		{
			std::vector<std::vector<unsigned int> > res(m_index.size());
			for(std::vector<int>::size_type i=0; i<m_index.size(); ++i)
			{
				res[m_index[i]].push_back(i);
			}
			std::vector<std::vector<unsigned int> >::iterator resIter(res.begin());
			while(resIter != res.end())
			{
				if(resIter->size()==0)
				{
					resIter = res.erase(resIter);
				}
				else
				{
				    ++resIter;
				}
			}
			return res;
		}
		
		/// Append \a set2 to \a set1.
		void unify(int set1, int set2)
		{
			if(set1 != set2)
			{
				for(std::vector<int>::size_type i=0; i<m_index.size(); ++i)
				{
					if(m_index[i] == set2)
					{
						m_index[i] = set1;
					}
				}
			}
		}
		
		/// Get the index of the set for position \a i.
		int getSetIndex(int i) const { return m_index[i]; }
		
		/// \a rows will be filled to the set with index \a setIndex.
		std::vector<unsigned int>& getRows(int setIndex, std::vector<unsigned int>& rows) const
		{
			rows.clear();	
			for(std::vector<int>::size_type i=0; i<m_index.size(); ++i)
			{
				if(m_index[i] == setIndex)
				{
					rows.push_back(i);
				}
			}
			return rows;
		}

		/// Count the non-empty sets.		
		int countSets(void) const;
		
		/// Get the sets with order > \a setsize.
		/// All the sets with index >= \a rank are of order 1.
		std::vector<unsigned int>& getBiggerSets(unsigned int rank, unsigned int setsize, std::vector<unsigned int>& indices) const; 
				
		/// Print the partition.		
		friend std::ostream & operator<< (std::ostream & os, const UnordPartition& p);
};

#endif
